"""
class HfstException and its subclasses

CLASSES:

    ContextTransducersAreNotAutomataException
    EmptySetOfContextsException
    EmptyStringException
    EndOfStreamException
    FlagDiacriticsAreNotIdentitiesException
    FunctionNotImplementedException
    HfstException
    HfstFatalException
    HfstTransducerTypeMismatchException
    ImplementationTypeNotAvailableException
    IncorrectUtf8CodingException
    MetadataException
    MissingOpenFstInputSymbolTableException
    NotTransducerStreamException
    NotValidAttFormatException
    NotValidLexcFormatException
    NotValidPrologFormatException
    SpecifiedTypeRequiredException
    StateIndexOutOfBoundsException
    StateIsNotFinalException
    StreamCannotBeWrittenException
    StreamIsClosedException
    StreamNotReadableException
    SymbolNotFoundException
    TransducerHasWrongTypeException
    TransducerHeaderException
    TransducerIsCyclicException
    TransducersAreNotAutomataException
    TransducerTypeMismatchException

"""

from libhfst import HfstException,\
HfstTransducerTypeMismatchException,\
ImplementationTypeNotAvailableException,\
FunctionNotImplementedException,\
StreamNotReadableException,\
StreamCannotBeWrittenException,\
StreamIsClosedException,\
EndOfStreamException,\
TransducerIsCyclicException,\
NotTransducerStreamException,\
NotValidAttFormatException,\
NotValidPrologFormatException,\
NotValidLexcFormatException,\
StateIsNotFinalException,\
ContextTransducersAreNotAutomataException,\
TransducersAreNotAutomataException,\
StateIndexOutOfBoundsException,\
TransducerHeaderException,\
MissingOpenFstInputSymbolTableException,\
TransducerTypeMismatchException,\
EmptySetOfContextsException,\
SpecifiedTypeRequiredException,\
HfstFatalException,\
TransducerHasWrongTypeException,\
IncorrectUtf8CodingException,\
EmptyStringException,\
SymbolNotFoundException,\
MetadataException,\
FlagDiacriticsAreNotIdentitiesException
